/*
 * tcp.c
 *
 * Copyright 2004, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * Transmission control Protocl.
 */


#include<types.h>
#include<lib.h>
#include<errno.h>
#include<mm.h>
#include<proc.h>
#include<time.h>
#include<net/net.h>
#include<net/ip.h>
#include<net/tcp.h>


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


enum{
	/* Tcp flag. */
	TCP_FLG_FIN=1<<0,
	TCP_FLG_SYN=1<<1,
	TCP_FLG_PST=1<<2,
	TCP_FLG_PSH=1<<3,
	TCP_FLG_ACK=1<<4,
	TCP_FLG_URG=1<<5,
};


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


/*
 * Note! nextǤƬ˻äƤ롣
 */
typedef struct RECV_BUF{
	struct RECV_BUF *next;
	uint             time;		/* Recieve time. */
	uint             srcip;		/* Source IP. */
	char             data[0];
}RECV_BUF;

typedef struct REPLY_WAIT{
	WAIT_INTR          wait;
	struct REPLY_WAIT *next;
	SOCKET            *sc;
}REPLY_WAIT;


/* PRIVATE */
static REPLY_WAIT *repWait=NULL;


/*
 * PRIVATE
 * Save receive data.
 * parameters : UDP data,source IP
 * return : 0 or task switch=1
 */
static int saveRecv(UDP_HEADER *head,uint srcip)
{
	int size;
	REPLY_WAIT *wait;
	RECV_BUF *recv;
	RECV_BUF *p;


	/* READԤץθ */
	wait=repWait;
	do
	{
		if(wait==NULL)return 0;
	}while(wait->sc->port!=swapWord(head->dstport));

	/* Save buffer. */
	size=swapWord(head->len);
	if((recv=kmalloc(sizeof(RECV_BUF)+size))==NULL)return 0;
	memcpy(recv->data,head,size);
	recv->srcip=srcip;
	recv->next=NULL;

	/* Add to buffer link. */
	if(wait->sc->buf==NULL)wait->sc->buf=recv;
	else
	{
		for(p=wait->sc->buf;p->next!=NULL;p=p->next)
		p->next=recv;
	}

	/* READԤץε */
	wake_intr(&wait->wait);

	return 1;
}


/*
 * PUBLIC
 * Get data from receive buffer.
 * parameters : buffer,buffer size,flags,SOCKET,source IP pointer
 * return : copy data size of error number
 */
static int getRecvBuf(void *buf,size_t len,int flags,SOCKET *sc,uint *srcip)
{
	enum{REPLY_TIMEOUT=2000};	/* ॢȥߥá */

	int size;
	int wf;
	REPLY_WAIT *wait=NULL;
	RECV_BUF *p;
	REPLY_WAIT *r;


	for(;;)
	{
		/* 쥷֥Хåե򸡺 */
		if(sc->buf!=NULL)
		{
			p=(RECV_BUF*)sc->buf;
			sc->buf=p->next;

			/* ǡХåե˥ԡ */
			size=((UDP_HEADER*)p->data)->len-sizeof(UDP_HEADER);
			if(size>len)size=len;
			memcpy(buf,((UDP_HEADER*)p->data)->data,size);
			*srcip=p->srcip;
			kfree(p);
			kfree(wait);

			return size;
		}

		/* ǡ쥷Ԥ */
		if((wait=kmalloc(sizeof(REPLY_WAIT)))==NULL)return -ENOMEM;
		wait->wait.flag=0;
		wait->sc=sc;
		wait->next=repWait;
		repWait=wait;
		wait_intr(&wait->wait,REPLY_TIMEOUT);

		/*  */
		wf=wait->wait.flag;
		if(repWait==wait)repWait=wait->next;
		else
			for(r=repWait;;r=r->next)
				if(r->next==wait)
				{
					r->next=wait->next;
					break;
				}
		kfree(wait);
		if(wf)return -EAGAIN;		/* Time out. */
	}
}


/*
 * GROBAL
 * Receive TCP frame.
 * parameters : IP data
 * return : 0 or task switch=1
 */
int receiveTcp(IP_HEADER *head)
{
	UDP_HEADER *tcp;


	tcp=(TCP_HEADER*)head->data;
/***************************************************************************************/
printk("RECV TCP srcport=%x,dstport=%x,length=%x\n",
	swapWord(udp->srcport),swapWord(udp->dstport),swapWord(udp->len));
/***************************************************************************************/
	/* åγǧ */
	if(calcSum((uint*)tcp,swapWord(tcp->len)))return 0;

	return saveRecv(tcp,head->srcip);
}


/**************************************************************************
 *
 * ͥ
 *
 **************************************************************************/


/*
 * PRIVATE
 * Get sequence number.
 * return : sequence number
 */
static inline uint32_t getSeqNumber()
{
	/* 4msǣ䤹RFC793ˡ */
	return rdtsc()/clock_1m*4;
}


/*
 * PUBLIC
 * Transmit SYN segment.
 */



/**************************************************************************
 *
 * Socket interface.
 *
 **************************************************************************/


static void send(SOCKET *sc,const void *msg,size_t len,int flags,struct sockaddr *to)
{
	UDP_HEADER udp;
	TRANS_BUF_INFO tbi;


	/* Make Header. */
	udp.srcport=swapWord(sc->port);
	udp.dstport=swapWord(((struct sockaddr_in*)to)->sin_port);
	udp.len=swapWord(len+sizeof(UDP_HEADER));
	udp.chksum=0;
	udp.chksum=calcSumM((uint*)&udp,(uint*)msg,sizeof(UDP_HEADER),len);

	/* Хåեơ֥ */
	tbi.data[2]=(char*)msg;
	tbi.size[2]=len;
	tbi.data[1]=(char*)&udp;
	tbi.size[1]=sizeof(UDP_HEADER);
	tbi.ipType=IPPROTO_UDP;
/********************************************************************************************
{
	TCP_HEADER *head=(TCP_HEADER*)msg;
	printk("TRANS UDP srcport=%x,dstport=%x,length=%x\n",udp->srcport,udp->dstport,udp->len);
}
*********************************************************************************************/
	transIp(&tbi,((struct sockaddr_in*)to)->sin_addr.s_addr,flags);
}


/*
 * return : copy data size or error number
 */
static int recv(SOCKET *sc,void *buf,size_t len,int flags,struct sockaddr *from,socklen_t *fromlen)
{
	int rest;
	uint srcip;


	if((rest=getRecvBuf(buf,len,flags,sc,&srcip))<0)return rest;
	if((from!=NULL)&&(fromlen!=NULL))
	{
		if((checkMem(from)==-1)||(checkMem(fromlen)==-1))return -EFAULT;
		if(*fromlen>=sizeof(struct sockaddr_in))
		{
			((struct sockaddr_in*)from)->sin_addr.s_addr=srcip;
			*fromlen=sizeof(struct sockaddr_in);
		}
	}

	return rest;
}


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


static SOCKET_INFO sockInfo={send,recv};


/*
 * GLOBAL
 * Init UDP.
 */
int initTcp()
{
	registSocket(&sockInfo,IPPROTO_TCP);

	return 0;
}
