/*
 * udp.c
 *
 * Copyright 2004, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * User datagram Protocl.
 */


#include<types.h>
#include<lib.h>
#include<errno.h>
#include<mm.h>
#include<proc.h>
#include<time.h>
#include<net/net.h>
#include<net/ip.h>
#include<net/ether.h>
#include<net/udp.h>


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


typedef struct{
	uint32_t srcip;
	uint32_t dstip;
	uint8_t  tmp;
	uint8_t  prot;
	uint16_t len;
}PSEUDO_HEADER;


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


/*
 * Note! nextǤƬ˻äƤ롣
 */
typedef struct RECV_BUF{
	struct RECV_BUF *next;
	uint             time;		/* Recieve time. */
	uint             srcip;		/* Source IP. */
	char             data[0];
}RECV_BUF;

typedef struct REPLY_WAIT{
	WAIT_INTR          wait;
	struct REPLY_WAIT *next;
	SOCKET            *sc;
}REPLY_WAIT;


/* PRIVATE */
static REPLY_WAIT *repWait=NULL;


/*
 * PRIVATE
 * Save receive data.
 * parameters : UDP data,source IP
 * return : 0 or task switch=1
 */
static int saveRecv(UDP_HEADER *head,uint srcip)
{
	int size;
	REPLY_WAIT *wait;
	RECV_BUF *recv;
	RECV_BUF *p;


	/* READԤץθ */
	wait=repWait;
	do
	{
		if(wait==NULL)return 0;
	}while(wait->sc->port!=swapWord(head->dstport));

	/* Save buffer. */
	size=swapWord(head->len);
	if((recv=kmalloc(sizeof(RECV_BUF)+size))==NULL)return 0;
	memcpy(recv->data,head,size);
	recv->srcip=srcip;
	recv->next=NULL;

	/* Add to buffer link. */
	if(wait->sc->buf==NULL)wait->sc->buf=recv;
	else
	{
		for(p=wait->sc->buf;p->next!=NULL;p=p->next)
		p->next=recv;
	}

	/* READԤץε */
	wake_intr(&wait->wait);

	return 1;
}


/*
 * PUBLIC
 * Get data from receive buffer.
 * parameters : buffer,buffer size,flags,SOCKET,source IP pointer
 * return : copy data size of error number
 */
static int getRecvBuf(void *buf,size_t len,int flags,SOCKET *sc,uint *srcip)
{
	enum{REPLY_TIMEOUT=2000};	/* ॢȥߥá */

	int size;
	int wf;
	REPLY_WAIT *wait=NULL;
	RECV_BUF *p;
	REPLY_WAIT *r;


	for(;;)
	{
		/* 쥷֥Хåե򸡺 */
		if(sc->buf!=NULL)
		{
			p=(RECV_BUF*)sc->buf;
			sc->buf=p->next;

			/* ǡХåե˥ԡ */
			size=swapWord(((UDP_HEADER*)p->data)->len)-sizeof(UDP_HEADER);
			if(size>len)size=len;
			memcpy(buf,((UDP_HEADER*)p->data)->data,size);
			*srcip=p->srcip;
			kfree(p);

			return size;
		}

		/* ǡ쥷Ԥ */
		if((wait=kmalloc(sizeof(REPLY_WAIT)))==NULL)return -ENOMEM;
		wait->wait.flag=0;
		wait->sc=sc;
		wait->next=repWait;
		repWait=wait;
		wait_intr(&wait->wait,REPLY_TIMEOUT);

		/*  */
		wf=wait->wait.flag;
		if(repWait==wait)repWait=wait->next;
		else
			for(r=repWait;;r=r->next)
				if(r->next==wait)
				{
					r->next=wait->next;
					break;
				}
		kfree(wait);
		if(wf)return -EAGAIN;		/* Time out. */
	}
}


/*
 * GROBAL
 * Receive UDP frame.
 * parameters : IP data
 * return : 0 or task switch=1
 */
int receiveUdp(IP_HEADER *head)
{
	UDP_HEADER *udp;
	PSEUDO_HEADER pshead;


	udp=(UDP_HEADER*)head->data;
/***************************************************************************************
printk("RECV UDP srcport=%x,dstport=%x,length=%x\n",
	swapWord(udp->srcport),swapWord(udp->dstport),swapWord(udp->len));
***************************************************************************************/
	/* åγǧ */
	pshead.srcip=head->srcip;
	pshead.dstip=head->dstip;
	pshead.tmp=0;
	pshead.prot=head->prot;
	pshead.len=udp->len;
	if(calcSumM((uint*)&pshead,(uint*)udp,sizeof(PSEUDO_HEADER),swapWord(udp->len)))return 0;

	return saveRecv(udp,head->srcip);
}


/**************************************************************************
 *
 * Socket interface.
 *
 **************************************************************************/


static void send(SOCKET *sc,const void *msg,size_t len,int flags,struct sockaddr *to)
{
	char head[sizeof(PSEUDO_HEADER)+sizeof(UDP_HEADER)];
	uint tmp;
	PSEUDO_HEADER *pshead;
	UDP_HEADER *udp;
	TRANS_BUF_INFO tbi;


	/* Set pseudo header. */
	pshead=(PSEUDO_HEADER*)head;
	pshead->dstip=((struct sockaddr_in*)to)->sin_addr.s_addr;
	pshead->srcip=ethDev[searchDevnum(pshead->dstip,&tmp)].ip;
	pshead->tmp=0;
	pshead->prot=IPPROTO_UDP;
	pshead->len=swapWord(len+sizeof(UDP_HEADER));
	
	/* Set UDP head. */
	udp=(UDP_HEADER*)(head+sizeof(PSEUDO_HEADER));
	udp->srcport=swapWord(sc->port);
	udp->dstport=swapWord(((struct sockaddr_in*)to)->sin_port);
	udp->len=swapWord(len+sizeof(UDP_HEADER));
	udp->chksum=0;
	udp->chksum=calcSumM((uint*)head,(uint*)msg,sizeof(PSEUDO_HEADER)+sizeof(UDP_HEADER),len);

	/* Хåեơ֥ */
	tbi.data[2]=(char*)msg;
	tbi.size[2]=len;
	tbi.data[1]=(char*)udp;
	tbi.size[1]=sizeof(UDP_HEADER);
	tbi.ipType=IPPROTO_UDP;
/********************************************************************************************
{
	UDP_HEADER *head=(UDP_HEADER*)msg;
	printk("TRANS UDP srcport=%x,dstport=%x,length=%x\n",udp->srcport,udp->dstport,udp->len);
}
*********************************************************************************************/
	transIp(&tbi,pshead->dstip,flags);
}


/*
 * return : copy data size or error number
 */
static int recv(SOCKET *sc,void *buf,size_t len,int flags,struct sockaddr *from,socklen_t *fromlen)
{
	int rest;
	uint srcip;


	if((rest=getRecvBuf(buf,len,flags,sc,&srcip))<0)return rest;
	if((from!=NULL)&&(fromlen!=NULL))
	{
		if((checkMem(from)==-1)||(checkMem(fromlen)==-1))return -EFAULT;
		if(*fromlen>=sizeof(struct sockaddr_in))
		{
			((struct sockaddr_in*)from)->sin_addr.s_addr=srcip;
			*fromlen=sizeof(struct sockaddr_in);
		}
	}

	return rest;
}


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


static SOCKET_INFO sockInfo={send,recv};


/*
 * GLOBAL
 * Init UDP.
 */
int initUdp()
{
	registSocket(&sockInfo,IPPROTO_UDP);

	return 0;
}
