/*
 * unistd.c
 *
 * Copyright 2003, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 */


#include<sys/types.h>
#include<stddef.h>
#include<callgate.h>
#include<share/syscall.h>
#include<system.h>
#include<errno.h>
#include<string.h>
#include<fcntl.h>
#include<sys/stat.h>
#include<signal.h>
#include<unistd.h>


/************************************************************************************
 *
 * exec
 *
 **************************************************************************************/

extern char **environ;


/*
 * PATHĶѿʸ롣
 */
static inline const char *getPathEnv()
{
	char **env=environ;
	char *pathstr="PATH=";
	int i;


	for(;env!=NULL;++env)
		for(i=0;(*env)[i]==pathstr[i];)
			if(pathstr[++i]=='\0')return &(*env)[i];

	return ":/ bin:/usr/bin";
}


/*
 * PATHĶѿʸ򥳥ԡ롣
 * return : pathμʸΥݥ(':' or '\0')
 */
static inline const char *cpyPathEnv(char *buf,const char *env)
{
	int i;


	for(i=0;env[i]!=':';++i)
		if((buf[i]=env[i])=='\0')return &env[i];
	buf[i]='\0';

	return &env[i];
}


int execve(const char *path, char *const argv[], char *const envp[])
{
	errno=-syscall4(SYS_EXEC,(int)path,(int)argv,(int)envp);
	return -1;
}


int execl(const char *path, const char *arg0, ...)
{
	return execve(path,(char**)&arg0,environ);
}


int execle(const char *path, const char *arg0, ...)
{
	char **env;


	env=(char**)&arg0;
	while(*env++!=NULL);
	return execve(path,(char**)&arg0,env);
}


int execv(const char *path, char *const argv[])
{
	return execve(path,argv,environ);
}


int execlp(const char *file, const char *arg0, ...)
{
	const char *epath;
	char buf[_PC_PATH_MAX];


	if(strchr(file,'/')==NULL)
	{
		/* pathĶѿõ */
		epath=getPathEnv();

		while(*epath!='0')
		{
			epath=cpyPathEnv(buf,epath);
			execve(strcat(buf,file),(char**)&arg0,environ);
			if(errno!=ENOENT)break;
			if(*epath==':')++epath;
		}

		return -1;
	}
	else return execve(file,(char**)&arg0,environ);
}


int execvp(const char *file, char *const argv[])
{
	const char *epath;
	char buf[_PC_PATH_MAX];


	if(strchr(file,'/')==NULL)
	{
		/* pathĶѿõ */
		epath=getPathEnv();

		while(*epath!='0')
		{
			epath=cpyPathEnv(buf,epath);
			execve(strcat(buf,file),argv,environ);
			if(errno!=ENOENT)break;
			if(*epath==':')++epath;
		}

		return -1;
	}
	else return execve(file,argv,environ);
}


/************************************************************************************
 *
 * process ID
 *
 ************************************************************************************/

pid_t getpid(void)
{
	PROC_ID pd;


	syscall3(SYS_GET_PROCID,0,(int)&pd);
	return pd.pid;
}


gid_t getgid(void)
{
	PROC_ID pd;


	syscall3(SYS_GET_PROCID,0,(int)&pd);
	return pd.gid;
}


uid_t getuid(void)
{
	PROC_ID pd;


	syscall3(SYS_GET_PROCID,0,(int)&pd);
	return pd.uid;
}


gid_t getegid(void)
{
	return getgid();
}


uid_t geteuid(void)
{
	return getuid();
}


pid_t getppid(void)
{
	PROC_ID pd;


	syscall3(SYS_GET_PROCID,0,(int)&pd);
	return pd.ppid;
}


pid_t getpgid(pid_t pid)
{
	int rest;
	PROC_ID pd;


	if((rest=syscall3(SYS_GET_PROCID,pid,(int)&pd))<0)
	{
		errno=-rest;
		return -1;
	}
	return pd.gid;
}


pid_t getpgrp(void)
{
	return getpgid(0);
}


int getgroups(int size, gid_t list[])
{
	/* ̤б */
	return 0;
}


int setgid(gid_t gid)
{
	int rest;
	PROC_ID pd;


	pd.gid=gid;
	pd.uid=-1;
	if((rest=syscall3(SYS_SET_PROCID,0,(int)&pd))<0)
	{
		errno=-rest;
		return -1;
	}
	return 0;
}


int setuid(uid_t uid)
{
	int rest;
	PROC_ID pd;


	pd.gid=-1;
	pd.uid=uid;
	if((rest=syscall3(SYS_SET_PROCID,0,(int)&pd))<0)
	{
		errno=-rest;
		return -1;
	}
	return 0;
}


int setegid(gid_t egid)
{
	return setgid(egid);
}


int setpgid(pid_t pid, pid_t pgid)
{
	int rest;
	PROC_ID pd;


	pd.gid=pgid;
	pd.uid=-1;
	if((rest=syscall3(SYS_SET_PROCID,pid,(int)&pd))<0)
	{
		errno=-rest;
		return -1;
	}
	return 0;
}


pid_t setsid(void)
{
	pid_t pid;


	pid=getpid();
	if(setgid(pid)==-1)return -1;
	return pid;
};


/************************************************************************************
 *
 * ¾
 *
 ************************************************************************************/

int chdir(const char *path)
{
	int rest;


	if((rest=syscall2(SYS_CHDIR,(int)path))<0)
	{
		errno=-rest;
		return -1;
	}

	return 0;
}


int dup(int fildes)
{
	return fcntl(fildes,F_DUPFD,0);
}


int dup2(int fildes,int fildes2)
{
	close(fildes2);
	return fcntl(fildes,F_DUPFD,fildes2);
}


void _exit(int state)
{
	syscall2(SYS_EXIT,(int)state);
}


pid_t fork(void)
{
	int rest;


	if((rest=syscall1(SYS_FORK))<0)
	{
		errno=-rest;
		return -1;
	}

	return rest;
}


char *getcwd(char *buf, size_t size)
{
	int rest;


	if((rest=syscall3(SYS_GETCWD,(int)buf,size))<0)
	{
		errno=-rest;
		return NULL;
	}

	return buf;
}


off_t lseek(int fildes,off_t offset,int whence)
{
	int rest;


	if((rest=syscall4(SYS_LSEEK,fildes,offset,whence))<0)
	{
		errno=-rest;
		return -1;
	}

	return rest;
}


ssize_t read(int fildes,void *buf,size_t nbyte)
{
	int rest;


	if((rest=syscall4(SYS_READ,fildes,(int)buf,(int)nbyte))<0)
	{
		errno=-rest;
		return -1;
	}

	return rest;
}


int rmdir(const char *path)
{
	int rest;


	if((rest=syscall2(SYS_RMDIR,(int)path))<0)
	{
		errno=-rest;
		return -1;
	}

	return 0;
}


ssize_t write(int fildes,const void *buf,size_t nbyte)
{
	int rest;


	if((rest=syscall4(SYS_WRITE,fildes,(int)buf,(int)nbyte))<0)
	{
		errno=-rest;
		return -1;
	}

	return rest;
}


int close(int fildes)
{
	int rest;


	/* ɸϤϥʤ */
	if((fildes>=STDIN_FILENO)&&(fildes<=STDERR_FILENO))
	{
		errno=EBADF;
		return -1;
	}

	if((rest=syscall2(SYS_CLOSE,fildes))<0)
	{
		errno=-rest;
		return -1;
	}

	return 0;
}


int brk(void *end_segment)
{
	static uint allocLast=0;
	uint alloc_last;


	if((alloc_last=ROUNDUP((uint)end_segment,PAGE_SIZE)-1)>allocLast)
	{
		if(syscall2(SYS_BRK,alloc_last)!=0)return -1;
		allocLast=alloc_last;
	}

	return 0;
}


void *sbrk(ptrdiff_t increment)
{
	void *begin;


	begin=(void*)(sys_info->lastAddr+1);

	if(increment>0)
	{
		if(brk((void*)(sys_info->lastAddr+increment))!=0)return NULL;
		sys_info->lastAddr+=increment;
	}

	return begin;
}


void sync()
{
	syscall1(SYS_SYNC);
}


int unlink(const char *path)
{
	int rest;


	if((rest=syscall2(SYS_UNLINK,(int)path))<0)
	{
		errno=-rest;
		return -1;
	}
	else return 0;
}


int chown(const char *path, uid_t owner, gid_t group)
{
	int rest=0;


	if((int)owner!=-1)
		rest=syscall4(SYS_CHATTR,(int)path,CHATTR_UID,(int)owner);
	if((int)group!=-1)
		rest=syscall4(SYS_CHATTR,(int)path,CHATTR_GID,(int)group);
	if((rest)<0)
	{
		errno=-rest;
		return -1;
	}
	else return 0;
}


int reboot(int flag)
{
	int rest;


	if((rest=syscall2(SYS_REBOOT,flag))<0)
	{
		errno=-rest;
		return -1;
	}
	else return 0;
}


unsigned int sleep(unsigned int seconds)
{
	return syscall2(SYS_SLEEP,seconds);
}


unsigned int alarm(unsigned int seconds)
{
	return syscall2(SYS_ALARM,seconds);
}


int pause(void)
{
	sigset_t mask=0;


	return sigsuspend(&mask);
}
